#!/usr/bin/env python
# -*- coding: utf8 -*-

import pathlib
import numpy as np
import pandas as pd
import xml.etree.ElementTree as ET
import lxml.etree
import lxml
import tqdm
import json
import glob
from mhealthdata.utils import *
from mhealthdata.df2numpy import _get_time, _get_idate_imin
from mhealthdata.df2numpy import to_1darray, to_2darray, combine_arrays
from mhealthdata.timezone import find_timezone_mismatch, fix_timezone_mismatch
import warnings
warnings.filterwarnings('ignore')


class DataLoader():
    """
    This is the class from which all loaders inherit.

    DataLoader subclasses make numpy arrays from basic health sensor data, 
    all at local time zone loaded from different mobile health apps.

    Basic health data are:
    - per minute values of "steps", "sleep", and "bpm"
    - per day values of "weight", "rhr", and "hrv"
    - per user values of "dob", "sex", and "height"

    Data can be accessed:
    - as pandas DataFrame in self.df dict attribute 
    - as numpy arrays using get_device_data() or save_device_npz() methods

    Methods are device-centric and can be used to get data for any specific device
    connected to your health app aggregater, e.g. iPhone, or Apple Watch, or Fitbit wristband

    Notes
    -----
    Other data can be found in data export (e.g. VO2Max), but not processed by DataLoader subclasses.
    In case those data needed, the self.category dict attribute should be modified.

    Parameters
    ----------
    path : str
        Path to unzipped local folder containing health app data

    Attributes
    ----------
    df : dict
        Dictionary of pandas DataFrames of loaded health data for "steps", "bpm", etc.
    categories : dict
        Dictionary of health data categories. Keys used to find files. Attributes: 
        "name" - to rename, "column" - to seek value column in corresponding DataFrame.
    userdata : dict
        Dictionary of "Date-of-birth", "Biological sex", and "Height".
        Other data like country of residence, or phone number, etc. are ignored.
    start_keys : list
        List of keywords to seek for start timestamp column in a DataFrame.
    end_keys : list
        List of keywords to seek for end timestamp column in a DataFrame.
    tz_keys : list
        List of keywords to seek for timezone in a DataFrame (NOT to be applied timestamps).
    tz_offset : list
        List of keywords to seek for timezone in a DataFrame (to be applied timestamps).
    dev_col : list
        List of keywords to seek for device id column in a DataFrame.
    path : list
        Path to unzipped local folder containing health app data.

    """

    def __init__(self, path):
        self.df = {}
        self.categories = {}
        self.userdata = {}
        self.start_keys = ["start_time", "startTime", "startDate", 
                           "dateTime", "day_time"]
        self.end_keys = ["end_time", "endTime", "endDate"]
        self.tz_keys = ["time_offset", "HKTimeZone"]
        self.tz_offset = ["time_offset"]
        self.dev_col = []
        self.path = path


    @property
    def devices_dict(self):
        """
        Get dictionary of loaded devices identifiers.

        Returns
        -------
        dict
            Dictionary of loaded devices identifiers.
        
        """
        return {"all": ["all"]}


    @property
    def devices(self):
        """
        Get list of loaded devices.

        Returns
        -------
        list
            List of loaded devices.
        
        """
        dev = self.devices_dict
        return list(dev.keys())


    @property
    def dataframes(self):
        """
        Get list of loaded DataFrames.
        
        Returns
        -------
        list
            List of loaded DataFrames.

        """
        return list(self.df.keys())


    @property
    def all_categories(self):
        """
        Get list of all data categories found (not all loaded).
        
        Returns
        -------
        list
            List of all data categories found in provided path.

        """
        categories = []
        return categories


    def _parse_userdata(self):
        """
        Private method to retrieve "Date-of-birth", "Biological sex", and "Height".
        Other data like country of residence, or phone number, etc. are ignored.

        Returns
        -------
        list
            List of loaded userdata keys.

        """
        return list(self.userdata.keys())


    @staticmethod
    def _special_cases(df, category):
        """
        Private method to process special cases during loading of data.
        
        Parameters
        ----------
        df : DataFrame
            DataFrames of loaded health data for "steps", "bpm", etc.
        category : str
            Key used to find health data files.

        Returns
        -------
        DataFrame
            DataFrame with applied health app-specific fixes.

        """
        return df


    @staticmethod
    def _get_duration(df):
        """
        Get record durations from loaded datframes.
        
        Notes
        -----
        "binning_period" is treated as [minutes]
        "stage_duration" and "seconds" are treated as [seconds]

        Parameters
        ----------
        df : DataFrame
            DataFrames of loaded health data for "steps", "bpm", etc.

        Returns
        -------
        ndarray
            1D array record durations [seconds]

        """
        dt = None
        if "binning_period" in df.columns:
            dt = 60 * df["binning_period"].values.astype(int)
        elif "seconds" in df.columns:
            dt = (df["seconds"].values + 1)
            dt = np.round(dt).astype(int)
        elif "stage_duration" in df.columns:
            dt = (df["stage_duration"].values + 1)
            dt = np.round(dt).astype(int)
        return dt


    @staticmethod
    def _get_device_slice(df, uuids, dev_col):
        """
        Private method to get DataFrame slice for specified device.
        
        Parameters
        ----------
        df : DataFrame
            DataFrames of loaded health data for "steps", "bpm", etc.
        uuids : list
            List of device identifiers.
        dev_col : list
            List of keywords to seek for device id column in a DataFrame.

        Returns
        -------
        DataFrame
            DataFrame slice matching mask of specified device identifiers.

        """
        mask = np.zeros((df.shape[0])).astype(bool)
        deviceuuid = find_columns_by_key(df, dev_col)
        if len(deviceuuid) > 0:
            deviceuuid = df[deviceuuid[0]].values.astype(str)
            for uuid in uuids:
                mask[deviceuuid == uuid] = True
        return df[mask]


    def _get_timezone(self, tstart, idate=None):
        """
        Search dataframes for timezone for provided ordinal dates.
        
        Parameters
        ----------
        tstart : list
            List of columns to seek for start date/time
        idate : ndarray or None, default None
            1D array of continuous range of ordinal days

        Returns
        -------
        ndarray
            1D array of timezone offset from GMT [minutes].

        """
        tz_dict = {}
        for category in self.categories:
            df = self.df[category]
            tz_col = find_columns_by_key(df, self.tz_keys)
            tz_col = tz_col[0] if len(tz_col) > 0 else None
            if tz_col is not None:
                t0 = _get_time(df, tstart)
                iday, imin = _get_idate_imin(t0)
                idate = idate if idate is not None else to_range(iday)
                tz = df[tz_col].values.astype(str)
                tz = timezone_txt_to_minutes(tz)
                for i in idate:
                    tz_list = tz_dict[i] if i in tz_dict else []
                    tz_list = tz_list + list(tz[iday == i])
                    tz_dict[i] = tz_list
        tz = np.zeros(len(idate),) * np.nan
        for k, i in enumerate(idate):
            tz_list = tz_dict[i] if i in tz_dict else []
            if len(tz_list) > 0:
                t = unique_sorted(tz_list)[0]
                tz[k] = t[0] if len(t) > 0 else np.nan
        return tz.astype(np.float16)


    def get_device_data(self, device="all", idate=None, uint8=True):
        """
        Get dictionary of per day and per minute ndarrays.
        The method is device-centric and can output data for specified device.
        
        Parameters
        ----------
        device : str, default "all"
            Device name (sould match any one of self.devices).
        date_range : ndarray or None, default None
            1D array of continuous range of ordinal days.
            If None, automatically get from on min and max dates of "steps".
        uint8 : bool, default True
            Flag to cast all health data to np.uint8 to save disk space.

        Returns
        -------
        dict
            Dictionary of ountput ndarrays.

        """
        data = {}
        if device not in self.devices:
            raise KeyError(f"Wrong device '{device}'. Use 'devices' property to get valid devices.")
        for category in self.categories:
            df = self.df[category]
            if category not in ["weight"] and device not in ["all"]:
                uuids = self.devices_dict[device]
                df = self._get_device_slice(df, uuids, self.dev_col)
            if df.shape[0] > 0:
                dt = self._get_duration(df)
                name = self.categories[category]["name"]
                column = self.categories[category]["column"]
                x = data[name] if name in data else None
                if name in ["weight", "rhr"]:
                    x, idate = to_1darray(df, column, self.start_keys, self.end_keys, self.tz_offset, idate, x, uint8)
                else:
                    mode = "count" if name == "steps" else "rate"
                    x, idate = to_2darray(df, column, self.start_keys, self.end_keys, self.tz_offset, dt, idate, x, uint8, mode)
                data[name] = x
        data["idate"] = idate
        data["tz"] = self._get_timezone(self.start_keys, idate)
        return data
    
    
    def save_device_npz(self, output_file, device="all", idate=None, uint8=True):
        """
        Save dictionary of per day and per minute ndarrays to npz.
        The method is device-centric and can output data for specified device.
        
        Parameters
        ----------
        output_file : str
            Path to output .npz file.
        device : str, default "all"
            Device name (sould match any one of self.devices).
        date_range : ndarray or None, default None
            1D array of continuous range of ordinal days.
            If None, automatically get from on min and max dates of "steps".
        uint8 : bool, default True
            Flag to cast all health data to np.uint8 to save disk space.

        Returns
        -------
        bool
            True if data ndarrays length not zero, else False.

        """

        data = self.get_device_data(device, idate, uint8)
        if data:
            np.savez_compressed(output_file, **data)
        return bool(len(data["idate"]))







class FitbitLoader(DataLoader):
    """
    
    Notes
    -----
    One may note that Fitbit exported
    
        - ``sleep`` in local time
        - ``steps`` and ``bpm`` timestamps in UTC 

    FitbitLoader

        - Attempts to infer time zone offset from data mismatch
        - Converts all timestamps to local time, see ``self._fix_timezone()``

    Example
    -------
    Assume we have data export ``MyFitbitData.zip`` downloaded to folder \
    ``/Users/username/Downloads/wearable_data/`` and unzipped into a subfolder \
    ``/Users/username/Downloads/wearable_data/User/``.

    >>> import mhealthdata
    >>> path = '/Users/username/Downloads/wearable_data/User/'
    >>> wdata = mhealthdata.FitbitLoader(path)
    
    """

    def __init__(self, path):
        super().__init__(path)
        self.categories = {
            "steps": {
                "name": "steps", "column": "value"},
            "sleep": {
                "name": "sleep", "column": "level"},
            "heart_rate": {
                "name": "bpm", "column": "value.bpm"},
            "resting_heart_rate": {
                "name": "rhr", "column": "value.value"},
            "weight": {
                "name": "weight", "column": "weight"},
        }
        self.load_data()


    @property
    def all_categories(self):
        fnames = glob.glob(self.path + "/*/*.csv")
        fnames += glob.glob(self.path + "/*/*.json")
        categories = []
        for fname in fnames:
            category = pathlib.Path(fname).stem.split("-")[0]
            if category not in categories:
                categories.append(category)
        return categories


    @staticmethod
    def _special_cases(df, category):
        if category == "weight":
            df["dateTime"] = df["date"] + " " + df["time"]
            df["weight"] = 0.4536 * df["weight"] # pounds to kg
        return df


    def _parse_userdata(self):
        try:
            fname = glob.glob(self.path + "/*/Profile.csv")[0]
        except IndexError as e:
            return None
        df= pd.read_csv(fname)
        df = self.df["Profile"] = df
        self.userdata["Date of birth"] = df["date_of_birth"].values.astype(str)[0]
        self.userdata["Biological sex"] = df["gender"].values.astype(str)[0]
        self.userdata["Height"] = df["height"].values.astype(str)[0]
        return list(self.userdata.keys())


    def get_device_data(self, device="all", idate=None, trunc=True):
        data = super().get_device_data(device, idate, trunc)
        if "sleep" in data and len(data["sleep"]) > 0:
            data["tz"] = find_timezone_mismatch(data)
            data = fix_timezone_mismatch(data, tz=data["tz"])
        return data
    

    def load_sleep(self):
        """
        Load sleep data from .json files.
        Path to seek files is taken from self.path attribute.
        
        Returns
        -------
        DataFrame
            DataFrame of raw data loaded from .csv.

        """
        fnamelist = glob.glob(self.path + "/*/" + "sleep" + "-*")
        if len(fnamelist) == 0:
            return None
        df = []
        for fname in tqdm.tqdm(fnamelist):
            for json_data in json.load(open(fname)):
                df.append(pd.json_normalize(json_data["levels"]["data"]))
                if "shortData" in json_data["levels"]:
                    df.append(pd.json_normalize(json_data["levels"]["shortData"]))
        df = pd.concat(df, ignore_index=True)
        return df


    def load_nonsleep(self, category):
        """
        Load non-sleep data from .json files.
        Path to seek files is taken from self.path attribute.
        
        Parameters
        ----------
        category : str
            Key used to find health data files.

        Returns
        -------
        DataFrame
            DataFrame of raw data loaded from .csv.

        """
        fnamelist = glob.glob(self.path + "/*/" + category + "-*")
        if len(fnamelist) == 0:
            return None
        df = [pd.json_normalize(json.load(open(fname))) for fname in tqdm.tqdm(fnamelist)]
        df = pd.concat(df, ignore_index=True) if len(df) > 0 else pd.DataFrame()
        return df


    def load_data(self):
        """
        Load data from .csv and .json files.
        Cycling over category from self.categories attribute.
        Path to seek files is taken from self.path attribute.
        
        Returns
        -------
        list
            List of loiaded DataFrames.

        """
        self.df = {}
        for category in list(self.categories.keys()):
            df = self.load_sleep() if category == "sleep" else self.load_nonsleep(category)
            if df is not None:
                df = self._special_cases(df, category)
                df = columns_to_datetime(df, self.start_keys, self.end_keys, self.tz_offset)
                self.df[category] = df
            elif "step" in category or "pedometer" in category:
                raise FileNotFoundError(f"Wrong 'path'. Cannot find files for '{category}'.")
        self._parse_userdata()
        return self.dataframes









class ShealthLoader(DataLoader):
    """
    
    Notes
    -----
    Samsung Health exports:
    
        - ``.json`` (``step`` binning data) timestamps in local time
        - ``.csv`` (``sleep``, ``bpm``, ``weight``) in UTC with \
            additional time zone column ``time_offset``

    ShealthLoader

        - Converts all timestamps to local time, see ``utils.columnscolumns_to_datetime()``

    Example
    -------
    Assume we have data export downloaded to folder  \
    ``/Users/username/Downloads/wearable_data/Samsung Health/`` \
    which contains a subfolder ``samsunghealth_<username>_<date-time>``.


    >>> import mhealthdata
    >>> path = '/Users/username/Downloads/wearable_data/Samsung Health/'
    >>> wdata = mhealthdata.ShealthLoader(path)
    
    """

    def __init__(self, path):
        super().__init__(path)
        self.start_keys = ["start_time"]
        self.dev_col = ["deviceuuid"]
        self.categories = {
            "pedometer_day_summary": {
                "name": "steps", "column": "mStepCount"},
            "step_daily_trend": {
                "name": "steps", "column": "count"},
            "heart_rate": {
                "name": "bpm", "column": "com.samsung.health.heart_rate.heart_rate"},
            "sleep": {
                "name": "sleep", "column": "stage"},
            "sleep_stage": {
                "name": "sleep", "column": "stage"},
            "weight": {
                "name": "weight", "column": "weight"},
        }
        self.load_data()



    @property
    def devices_dict(self):
        def clean_username(s):
            s = " ".join(w for w in s.split() if "'" not in w)
            s = " ".join(w for w in s.split() if "(" not in w)
            return s
        dev = {}
        if "device_profile" in self.df:
            col = ["deviceuuid", "device_group", "name", "model", "fixed_name"]
            df = self.df["device_profile"][col].astype(str)
            dev_group = df["device_group"].values.astype(int).clip(360001)
            dev_uuid = df["deviceuuid"].values.astype(str)
            dev_name = df["fixed_name"].values.astype(str)
            dev_name = np.where(dev_name == "nan", df["model"], dev_name)
            dev_name = np.where(dev_group == 360003, df["name"], dev_name)
            dev_name = [clean_username(name) for name in dev_name]
            dev_name = np.array(dev_name, dtype=str)
            dev["all"] = dev_uuid
            if 360001 in dev_group:
                dev["mobile"] = dev_uuid[dev_group == 360001]
            if 360003 in dev_group:
                dev["wearable"] = dev_uuid[dev_group == 360003]
            for name in np.unique(dev_name):
                dev[name] = dev_uuid[dev_name == name]
        return dev

    
    @property
    def all_categories(self):
        fnames = glob.glob(self.path + "/*/*.csv")
        categories = []
        for fname in fnames:
            category = fname.split(".")
            if len(category) > 3:
                categories.append(category)
        return categories


    @staticmethod
    def _special_cases(df, category):
        if "sleep" in category:
            if "stage" in df.columns:
                s = {40001: "awake", 40002: "light", 40003: "deep", 40004: "rem"}
                df["stage"] = np.vectorize(s.get)(df["stage"].values)
            else:
                df["stage"] = np.array(["no_stage"] * df.shape[0])
        return df
            

    def _parse_userdata(self):
        if "user_profile" in self.df:
            d = self.df["user_profile"]
            d["text_value"].fillna(d["float_value"], inplace=True)
            d.set_index("key", inplace=True)
            d = d[["text_value"]].T.to_dict("list")
            self.userdata["Date of birth"] = d["birth_date"][0]
            self.userdata["Biological sex"] = d["gender"][0]
            self.userdata["Height"] = d["height"][0]
        return list(self.userdata.keys())
    

    def _binning_dict(self, category, idx):
        """
        Private method to get device-id and date dictionaries for daily .json.
        
        Parameters
        ----------
        category : str
            Key used to find health data files.
        idx : str
            Column name containing binning data file names.

        Returns
        -------
        dev : dict
            Dictionary of device id for binning data file names.
        dat : dict
            Dictionary of dates for binning data file names.

        """
        df = self.load_csv(category)
        df.set_index(idx, inplace=True)
        dat = df["day_time"].to_dict()
        dev = df["deviceuuid"].to_dict()
        return dev, dat


    def load_csv(self, category):
        """
        Load data from .csv file.
        Path to seek files is taken from self.path attribute.
        
        Parameters
        ----------
        category : str
            Key used to find health data files.

        Returns
        -------
        DataFrame
            DataFrame of raw data loaded from .csv.

        """
        try:
            fname = glob.glob(self.path + "/*/*." + category + ".*.csv")[0]
        except IndexError as e:
            return None
        df = pd.read_csv(fname, skiprows=1, index_col=False)
        return df


    def load_jsons(self, category, idx="binning_data"):
        """
        Load data from .json files.
        Path to seek files is taken from self.path attribute.
        
        Parameters
        ----------
        category : str
            Key used to find health data files.
        idx : str, default "binning_data"
            Column name containing binning data file names.

        Returns
        -------
        DataFrame
            DataFrame of raw data loaded from .csv.

        """
        dev, dat = self._binning_dict(category, idx)
        fnamelist = glob.glob(self.path + '/*/jsons/*' + category + '*/*/*.' + idx + '.json')
        if len(fnamelist) == 0:
            return None
        df = []
        for fname in tqdm.tqdm(fnamelist):
            df_ = pd.json_normalize(json.load(open(fname)))
            nrec, ncol = df_.shape
            if ncol > 0:
                f = fname.split("/")[-1]
                df_["deviceuuid"] = np.array([dev[f]] * nrec)
                df_["start_time"] = dat[f] + 600000 * df_.index
                df_["binning_period"] = 10 * np.ones((nrec)).astype(int)
                df.append(df_)
        df = pd.concat(df, ignore_index=True)
        return df


    def load_data(self):
        """
        Load data from .csv and .json files.
        Cycling over category from self.categories attribute.
        Path to seek files is taken from self.path attribute.
        
        Returns
        -------
        list
            List of loiaded DataFrames.

        """
        self.df = {}
        for category in list(self.categories.keys()) + ["device_profile", "user_profile"]:
            df = self.load_csv(category)
            if "binning_data" in df.columns:
                df = self.load_jsons(category)
            if df is not None:
                df = self._special_cases(df, category)
                df = columns_to_datetime(df, self.start_keys, self.end_keys, self.tz_offset)
                self.df[category] = df
            elif "step" in category or "pedometer" in category:
                raise FileNotFoundError(f"Wrong 'path'. Cannot find files for '{category}'.")
        self._parse_userdata()
        return self.dataframes


    






class HealthkitLoader(DataLoader):
    """

    Example
    -------
    Assume path contains unzipped data ``export.xml`` or ``exportación.xml``.


    >>> import mhealthdata
    >>> path = '/Users/username/Downloads/wearable_data/apple_health_export/'
    >>> wdata = mhealthdata.HealthkitLoader(path)
    
    """

    
    def __init__(self, path):
        super().__init__(path)
        self.dev_col = ["sourceName"]
        self.categories = {
            "HKQuantityTypeIdentifierStepCount": {
                "name": "steps", "column": "value"},
            "HKCategoryTypeIdentifierSleepAnalysis": {
                "name": "sleep", "column": "value"},
            "HKQuantityTypeIdentifierHeartRate": {
                "name": "bpm", "column": "value"},
            "HKQuantityTypeIdentifierRestingHeartRate": {
                "name": "rhr", "column": "value"},
            "HKQuantityTypeIdentifierHeartRateVariabilitySDNN": {
                "name": "hrv", "column": "value"},
            "HKQuantityTypeIdentifierBodyMass": {
                "name": "weight", "column": "value"},
        }
        self.load_data()


    @property
    def devices_dict(self):
        dev = {"all": ["all"]}
        for category in self.categories:
            if category in self.df:
                df = self.df[category]
                if "sourceName" in df.columns:
                    for d in np.unique(df["sourceName"].values.astype(str)):
                        dev[d] = [d]
        return dev


    @property
    def all_categories(self):
        categories = []
        for tag in ["Record", "Workout"]:
            df = self.df[tag]
            col = find_columns_by_key(df, ["type"])
            if len(col) > 0:
                categories.append(df[col[0]].values.astype(str))
        categories = np.concatenate(categories)
        categories = np.unique(categories).tolist()
        return categories


    @staticmethod
    def _special_cases(df, category):
        def clean_username(s):
            if "iphone" in s.lower():
                s = "iPhone"
            elif "apple" in s.lower() and "watch" in s.lower():
                s = "Apple Watch"
            return s
        if "sourceName" in df.columns:
            df["sourceName"] = df["sourceName"].apply(clean_username)
        if category == "HKCategoryTypeIdentifierSleepAnalysis":
            s = {"HKCategoryValueSleepAnalysisAwake": "awake", 
                 "HKCategoryValueSleepAnalysisInBed": "no_stage", 
                 "HKCategoryValueSleepAnalysisAsleep": "asleep"}
            df["value"] = np.vectorize(s.get)(df["value"].values)
        return df

    def _parse_userdata(self, data):
        if "Me" in data:
            dob = data["Me"]["HKCharacteristicTypeIdentifierDateOfBirth"]
            sex = data["Me"]["HKCharacteristicTypeIdentifierBiologicalSex"]
            self.userdata["Date of birth"] = dob.values.astype(str)[0]
            self.userdata["Biological sex"] = sex.values.astype(str)[0]
        if "Record" in data:
            mask = data["Record"]["type"] == "HKQuantityTypeIdentifierHeight"
            height = data["Record"][mask]
            height = height["value"] + " " + height["unit"]
            self.userdata["Height"] = unique_sorted(height.values.astype(str))[0][0]
        return list(self.userdata.keys())


    def _parse_xml(self, root):
        """
        Private method to parse records from loaded .xml
        Path to seek files is taken from self.path attribute.
        
        Parameters
        ----------
        root
            Root element attribute for .xml tree

        Returns
        -------
        dict
            Dictionary of DataFrames.

        """
        data = {}
        for tag in ["Record", "Workout", "Me"]:
            records = []
            for child in tqdm.tqdm(root):
                if child.tag == tag:
                    for node in list(child):
                        if node.tag == "MetadataEntry":
                            if node.attrib["key"] == "HKTimeZone":
                                child.attrib["HKTimeZone"] = node.attrib["value"]
                    records.append(dict(child.attrib))
            df = pd.DataFrame(records)
            df = columns_to_datetime(df, self.start_keys, self.end_keys, self.tz_offset)
            data[tag] = df
        return data


    def load_data(self):
        """
        Load data from .xml file.
        Cycling over category from self.categories attribute.
        Path to seek files is taken from self.path attribute.
        
        Returns
        -------
        list
            List of loiaded DataFrames.

        """
        try:
            fname = (glob.glob(self.path + "/[eE][xX][pP][oO][rR][tT].[xX][mM][lL]") + \
            glob.glob(self.path + "/[eE][xX][pP][oO][rR][tT][aA][cC][iI]*[nN].[xX][mM][lL]"))[0]
        except IndexError as e:
            raise FileNotFoundError(f"Wrong 'path'. Cannot find file 'export.xml'.")
        parser = lxml.etree.XMLParser(recover=True)
        tree = ET.parse(fname, parser=parser)
        root = tree.getroot()
        data = self._parse_xml(root)
        for tag in data:
            self.df[tag] = data[tag]
        self._parse_userdata(data)
        for category in self.categories:
            mask = data["Record"]["type"] == category
            df = data["Record"][mask].copy()
            df = self._special_cases(df, category)
            self.df[category] = df
        return self.dataframes
        



import types
__all__ = [name for name, thing in globals().items()
          if not (name.startswith('_') or isinstance(thing, types.ModuleType))]
del types




#!/usr/bin/env python
# -*- coding: utf8 -*-

import numpy as np
import pandas as pd
from mhealthdata.utils import *


def _get_values(df, column):
    """
    Get record value arrays from loaded datframes.
    
    Parameters
    ----------
    df : DataFrame
        DataFrame of loaded health data for "steps", "sleep", "bpm", etc.
    column : str
        Values column name

    Returns
    -------
    ndarray
        1D array of record values, e.g. sleep stage, or stepcounts, etc.

    """
    values = df[column].values
    try:
        values = values.astype(float)
    except:
        s = sleep_stage_dict(mode="encode")
        s["None"] = 0
        values = values.astype(str)
        values = np.vectorize(s.get)(values)
        values = values.astype(float)
    return values


def _get_time(df, keys):
    """
    Get record timestamps from loaded datframes.
    
    Parameters
    ----------
    df : DataFrame
        DataFrames of loaded health data for "steps", "bpm", etc.
    keys : list
        List of keywords to seek for a column in a DataFrame

    Returns
    -------
    ndarray
        1D array of type datetime64

    """
    col = find_columns_by_key(df, keys)
    t = df[col[0]] if len(col) > 0 else None
    return t


def _get_idate_imin(t):
    """
    Get record ordinal days and minutes from loaded datframes.
    
    Parameters
    ----------
    t : ndarray
        1D array of datetime64

    Returns
    -------
    idate : ndarray
        1D array of ordinal days (January 1 of year 1 is day 1)
    imin : ndarray
        1D array of minutes since midnight

    """
    idate = t.apply(pd.Timestamp.toordinal).values.astype(int)
    imin = (60 * t.dt.hour + t.dt.minute).values.astype(int)
    return idate, imin


def _calc_duration(t0, t1):
    """
    Private method to calc record durations from start and end timestamps.
    
    Parameters
    ----------
    t0 : ndarray
        1D array of start datetime64
    t1 : ndarray
        1D array of end datetime64; if None, set t1 = t0 + 1 [min]

    Returns
    -------
    ndarray
        1D array record durations [minutes]

    """
    dt = np.ones((len(t0))).astype(int)
    if t1 is not None:
        dt = (t1 - t0).astype("timedelta64[s]").values
        dt = (dt / 60).astype(int)
        dt[dt < 1] = 1
    return dt


def to_1darray(df, column, tstart, tend=None, tz=None, idate=None, x=None, uint8=False):
    """
    Get value-per-day health data ("weight", "rhr", or "hrv") as 1D array.
    
    Parameters
    ----------
    df : DataFrame
        DataFrames of health data records - "steps", "bpm", etc.
    column : str
        Name of values column
    tstart : list
        List of columns to seek for start date/time
    tend : list or None, default None
        List of columns to seek for end date/time
    tz : list or None, default None
        List of columns to seek for date/time time zone
    idate : ndarray or None, default None
        1D array of continuous range of ordinal days
    x : ndarray or None, default None
        Initialized 1D array; if None, will be initialized with np.zeros()
    uint8 : bool, default False
        Flag to cast all health data to np.uint8 to save disk space

    Returns
    -------
    x : ndarray
        1D array of values of size (N days).
    idate : ndarray
        1D array of record ordinal days.

    """
    df = columns_to_datetime(df, tstart, tend, tz)
    val = _get_values(df, column)
    t0 = _get_time(df, tstart)
    iday, imin = _get_idate_imin(t0)
    idate = idate if idate is not None else to_range(iday)
    n = len(idate)
    x = np.zeros((n)) if x is None else x
    idx = iday - idate[0]
    mask = (idx >= 0) & (idx < n)
    idx = idx[mask]
    val = val[mask]
    for k in range(len(val)):
        i = idx[k]
        x[i] = val[k]
    if uint8:
        x = np.clip(x,0,255).astype(np.uint8)
    return x, idate


def to_2darray(df, column, tstart, tend=None, tz=None, dt=None, idate=None, x=None, uint8=False, mode="rate"):
    """
    Get value-per-minute health data ("steps", "sleep", or "bpm") as 2D array.
    
    Parameters
    ----------
    df : DataFrame
        DataFrames of health data records - "steps", "bpm", etc.
    column : str
        Name of values column
    tstart : list
        List of columns to seek for start date/time
    tend : list or None, default None
        List of columns to seek for end date/time
    tz : list or None, default None
        List of columns to seek for date/time time zone
    dt : str, ndarray, or None, default None
        Column name or 1D array of record durations [seconds].
    idate : ndarray or None, default None
        1D array of continuous range of ordinal days
    x : ndarray or None, default None
        Initialized 1D array; if None, will be initialized with np.zeros()
    uint8 : bool, default False
        Flag to cast all health data to np.uint8 to save disk space
    mode : {"rate", "count"}, default "rate"
        Way to treat values of records longer than 1 minute: 
        if "rate" - duplicate values, if "count" - split evenly between minutes

    Returns
    -------
    x : ndarray
        2D array of values of size (N days x 1440 minutes).
    idate : ndarray
        1D array of record ordinal days.

    """
    assert mode in ["rate", "count"]
    df = columns_to_datetime(df, tstart, tend, tz)
    val = _get_values(df, column)
    t0, t1 = [_get_time(df, k) for k in [tstart, tend]]
    iday, imin = _get_idate_imin(t0)
    dt = df[dt].values if isinstance(dt, str) else dt # seconds
    dt = dt / 60 if dt is not None else _calc_duration(t0, t1) # minutes
    dt = np.ceil(dt).astype(int)
    idate = idate if idate is not None else to_range(iday)
    n = 1440 * len(idate)
    x = np.zeros((n)) if x is None else x.flatten()
    idx = 1440 * (iday - idate[0]) + imin
    mask = (idx >= 0) & (idx < n)
    idx = idx[mask]
    val = val[mask]
    for k in range(len(val)):
        i = idx[k]
        j = i + dt[k]
        if mode == "count":
            x[i:j] = val[k] / dt[k]
        else:
            x[i:j] = val[k]
    x = x.reshape(-1,1440)
    if uint8:
        x = np.clip(x,0,255).astype(np.uint8)
    return x, idate


def combine_arrays(*args, labels=None, mode="valid"):
    """
    Convert arrays of e.g. steps, bpm, sleep into 
    the same length and combine in a dictionary
    
    Parameters
    ----------
    *args
        Tuples of (data, date) e.g. as output by to_2darray()
    labels : list or None, default None
        List of keyword labels for data
    mode : {"valid", "full"}, default "valid"
        If "valid" all arrys shrinked to min overlapping range, else expanded
        
    Returns
    -------
    dict
        Dictionary with numpy arrays of the same length
    
    """
    assert labels is None or len(labels) == len(args)
    assert mode in ["full", "valid"]
    data = {}
    if mode == "valid":
        t0 = max(a[1][0] for a in args)
        t1 = min(a[1][-1] for a in args) + 1
        data["idate"] = np.arange(t0,t1)
        for i in range(len(args)):
            label = f"x{i}" if labels is None else labels[i]
            mask = (args[i][1] >= t0) & (args[i][1] < t1)
            data[label] = args[i][0][mask]
    if mode == "full":
        t0 = min(a[1][0] for a in args)
        t1 = max(a[1][-1] for a in args) + 1
        data["idate"] = np.arange(t0,t1)
        for i in range(len(args)):
            x, t = args[i]
            label = f"x{i}" if labels is None else labels[i]
            val = np.zeros((t1-t0,1440), x.dtype)
            i0 = t[0] - t0
            i1 = i0 + len(t)
            val[i0:i1] = x
            data[label] = val
    return data
        






#!/usr/bin/env python
# -*- coding: utf8 -*-

import numpy as np
from mhealthdata.utils import *


def _get_tzlist(dt):
    """
    Get array of possible timezone offsets

    Parameters
    ----------
    dt : int, default 60
        Stride to find best match timezone offset, [minutes]

    Returns
    -------
    ndarray
        1D array of possible timezone offsets

    """
    n = 1440 // dt
    tzlist = (np.arange(n) - n // 2) * dt
    return tzlist
    

def _get_slice(x, y, i, nday=3, backwards=True):
    """
    Get window slice

    Parameters
    ----------
    x : ndarray
        2D array of size N days x 1440 minutes (padded sleep > 0)
    y : ndarray
        2D array of size N days x 1440 minutes (padded steps + bpm, 0 replaced by np.nan)
    i : int
        Day index
    nday : int, default 3
        Window to find best match timezone offset, [days]
    backwards : bool, default True
        If True include previous day, otherwise include following days

    Returns
    -------
    ndarrays
        1D arrays of window slices

    """
    i0 = i - nday * backwards
    i1 = i0 + nday + 1
    return x[i0:i1].flatten(), y[i0:i1].flatten()


def _conv_slice(x, y, t):
    """
    Calculate convolution of window slices

    Parameters
    ----------
    x : ndarray
        2D array of size N days x 1440 minutes (padded sleep > 0)
    y : ndarray
        2D array of size N days x 1440 minutes (padded steps + bpm, 0 replaced by np.nan)
    t : int
        Offset [minutes]

    Returns
    -------
    float
        Convolution (the smaller - the better sleep-to-steps match)

    """
    mask = np.roll(x, -t) > 0
    return np.nanmean(y[mask]) * np.sum(mask)


def _calc_timezone(x, y, i, nday=3, dt=60):
    """
    Day-wise score for each timezone offset

    Parameters
    ----------
    x : ndarray
        2D array of size N days x 1440 minutes (padded sleep > 0)
    y : ndarray
        2D array of size N days x 1440 minutes (padded steps + bpm, 0 replaced by np.nan)
    i : int
        Day index
    nday : int, default 3
        Window to find best match timezone offset, [days]
    dt : int, default 60
        Stride to find best match timezone offset, [minutes]

    Returns
    -------
    ndarray
        1D array of score for each timezone offset

    """
    tzlist = _get_tzlist(dt)
    x_, y_ = _get_slice(x, y, i + nday, nday, backwards=False)
    score = np.array([_conv_slice(x_, y_, t) for t in tzlist])
    x_, y_ = _get_slice(x, y, i + nday, nday, backwards=True)
    score_ = np.array([_conv_slice(x_, y_, t) for t in tzlist])
    score = np.nanmin(np.stack([score, score_]), axis=0)
    score = score - np.nanmin(score)
    score[score > 0.1 * np.nanmax(score)] = np.nan
    mask = score == np.nanmin(score)
    t = tzlist[mask][0] if any(mask) else np.nan
    return score


def _baseline_timezone(tz, nday, dt):
    """
    Find baseline (i.e. long-term) timezone offsets

    Parameters
    ----------
    tz : ndarray
        1D array of length N days with timezone offset [minutes]
    nday : int, default 3
        Window to find best match timezone offset, [days]
    dt : int, default 60
        Stride to find best match timezone offset, [minutes]

    Returns
    -------
    ndarray
        1D array of length N days with baseline offset [minutes]

    """
    bz = np.zeros_like(tz) * np.nan
    tzlist = np.unique(tz[np.isfinite(tz)])
    for t in tzlist:
        idx = find_intervals(tz == t, tol = 4 * nday)
        for i0, i1 in idx:
            if i1 - i0 >= 4 * nday:
                bz[i0:i1] = t
    mask = np.isfinite(bz)
    if any(mask):
        idx = np.arange(len(bz))[mask]
        bz[:idx.min()] = bz[idx.min()]
        bz[idx.max():] = bz[idx.max()]
        idx = find_intervals(np.isnan(bz))
        if any(np.isfinite(bz)):
            for i0, i1 in idx:
                tx, _ = unique_sorted(tz[i0:i1])
                bx = np.zeros((2)) * np.nan
                bx[0] = bz[i0 - 1] if i0 > 0 else bz[i1]
                bx[1] = bz[i1] if i1 < len(bz) - 1 else bz[i0 - 1]
                dx = np.abs(bx - tx[0])
                if any(np.isfinite(dx)):
                    bz[i0:i1] = bx[np.nanargmin(dx)]
                else:
                    bz[i0:i1] = bx[np.isfinite(bx)][0]
    return bz


def _filter_timezone(tz, nday, dt):
    """
    Filter out too short or ambigouos timezone offsets

    Parameters
    ----------
    tz : ndarray
        1D array of length N days with timezone offset [minutes]
    nday : int, default 3
        Window to find best match timezone offset, [days]
    dt : int, default 60
        Stride to find best match timezone offset, [minutes]

    Returns
    -------
    ndarray
        1D array of length N days with timezone offset [minutes]

    """
    tz = tz.astype(float)
    if any(np.isfinite(tz)):
        bz = _baseline_timezone(tz, nday, dt)
        dz = tz - bz
        idx = find_intervals(dz != 0, tol=1)
        for i0, i1 in idx:
            n = max(np.sum(dz[i0:i1] >= 0), np.sum(dz[i0:i1] <= 0))
            if i1 - i0 <= nday or i1 - i0 > n:
                tz[i0:i1] = bz[i0:i1]
            else:
                tz[i0:i1] = unique_sorted(tz[i0:i1])[0][0]
    return tz


def _localize_activity(x, idx, tz):
    """
    Roll activity (steps or bpm) to local time zone

    Parameters
    ----------
    x : ndarray
        2D array of size N days x 1440 minutes
    idx : ndarray
        2D array of size N intervals x 2 indices (start, end)
    tz : ndarray
        1D array of length N intervals with timezone offset, [minutes]

    Returns
    -------
    ndarray
        2D array of size N days x 1440 minutes

    """
    x_ = np.zeros((x.size + 2 * 1440))
    for k, i in enumerate(idx):
        dt = tz[k]
        j = 1440 + i + dt
        x_[j[0]:j[1]] = x[i[0]:i[1]]
    x_ = x_[1440:-1440].reshape(-1,1440)
    return x_


def find_timezone_mismatch(data, nday=3, dt=60):
    """
    For Fitbit only: Identify timezone by mismatch of sleep, steps, and bpm
    - Assume "sleep" is in local time
    - Automatically detect day-wise "steps" & "bpm" offsets to match "sleep"
    
    Parameters
    ----------
    data : dict
        Dictionary of data, should have keys "sleep", "steps", "bpm", 
        each containing array of size N days x 1440 minutes
    nday : int, default 3
        Window to find best match timezone offset, [days]
    dt : int, default 60
        Stride to find best match timezone offset, [minutes]

    Returns
    -------
    ndarray
        1D array of length N days with timezone offset [minutes]

    """
    pad = np.zeros((nday,1440))
    x = np.vstack([pad, data["sleep"], pad])
    y = np.vstack([pad, data["steps"], pad]) + np.vstack([pad, data["bpm"], pad])
    x = (x > 0).astype(float)
    y[y == 0] = np.nan
    score = []
    for i in range(len(data["steps"])):
        score.append(_calc_timezone(x, y, i, nday, dt))
    score = np.stack(score)
    tzlist = _get_tzlist(dt)
    tz = np.zeros(len(score),) * np.nan
    for k, s in enumerate(score):
        if any(np.isfinite(s)):
            tz[k] = tzlist[np.nanargmin(s)]
    tz[0] = tz[1]
    tz[-1] = tz[-2]
    tz = _filter_timezone(tz, nday, dt)
    return tz

    
    
def fix_timezone_mismatch(data, tz=None):
    """
    Fix timezone offset for data imported from Fitbit
    - Assume "sleep" is in local time
    - Automatically detect day-wise "steps" and "bpm" offsets to match "sleep"
    
    Parameters
    ----------
    data : dict
        Dictionary of data, should have keys "sleep", "steps", "bpm"
        Each key should point to an array of size N days x 1440 minutes
    tz : ndarray or None, default None
        1D array of length N days with timezone offset [minutes]
        If None, timezone will be detected automatically

    Returns
    -------
    dict
        Dictionary of data with "steps" and "bpm" rolled to match "sleep"

    """
    tz = tz if tz is not None else find_timezone_mismatch(data)
    tz = np.repeat(tz, 1440)
    mask = np.diff(tz, prepend=tz[0]+1, append=tz[-1]+1) != 0
    idxs = find_intervals(data["sleep"].flatten())
    for k, (i0,i1) in enumerate(idxs):
        dt = unique_sorted(tz[i0:i1])[0]
        dt = int(dt[0]) if len(dt) > 0 else 0
        j0 = max(0, i0 - dt * (dt < 0))
        j1 = min(len(tz), i1 - dt * (dt > 0))
        dts = np.unique(tz[j0:j1])
        tz[j0:j1] = dt
    mask = np.diff(tz, prepend=tz[0]+1, append=tz[-1]+1) != 0
    idx = np.arange(len(tz) + 1)[mask]
    idx = np.stack([idx[:-1], idx[1:]]).T
    tz = np.array([np.unique(tz[i0:i1])[0] for i0, i1 in idx]).astype(int)
    data["steps"] = _localize_activity(data["steps"].flatten(), idx, tz)
    data["bpm"] = _localize_activity(data["bpm"].flatten(), idx, tz)
    return data





#!/usr/bin/env python
# -*- coding: utf8 -*-

import numpy as np
import pandas as pd
import pylab as plt
import itertools
import datetime
import calendar
import pytz
import re


def sleep_stage_dict(mode="decode"):
    """
    Gets dictionary to encode/decode sleep stage.
    
    Parameters
    ----------
    mode : {"decode", "encode"}, default "decode"
        If "decode", return dict num -> str
        If "encode", return dict str -> num

    Returns
    -------
    dict
        Dictionary for decoding/encoding sleep stages

    """
    assert mode in ["decode", "encode"]
    d = {
        "no_sleep": 0,
        "wake": 1,
        "awake": 1,
        "unknown": 2,
        "no_stage": 2,
        "rem": 3,
        "restless": 4,
        "light": 5,
        "asleep": 6,
        "deep": 7,
    }
    if mode == "decode":
        d = {val: key for key, val in d.items()}
    return d


def find_columns_by_key(df, keys):
    """
    Finds all DataFrame column names containing any of the keys.
    
    Parameters
    ----------
    df : DataFrame
        DataFrame
    keys : array_like
        List of keys of `str` type

    Returns
    -------
    list
        List of column names

    """
    keys = keys if keys is not None else []
    keys = keys if isinstance(keys, list) else [keys]
    col = [c for c in df.columns if any([k.lower() in c.lower() for k in keys])]
    return col


def find_intervals(x, tol=0, nmin=None, sort=False):
    """
    Finds continuous positive intervals in 1d-array.
    
    Parameters
    ----------
    x : ndarray
        1D array of non-negative numeric values
    tol : int, default 0
        Gap duration tolerance
    nmin : int or None, default None
        Minimal length of intervals
    sort : bool, default False
        If False - keep interval order by index in the array 
        If True - sort descending by interval duration

    Returns
    -------
    ndarray
        2D array - N intervals x 2 indices (start, end)

    """
    assert x.ndim == 1 and not any(x < 0)
    x_ = fill_gaps(x, tol)
    x_ = (x_ > 0).astype(int)
    pad = np.zeros((1))
    x_ = np.concatenate([pad, x_, pad])
    x_ = np.diff(x_)
    idx = np.arange(len(x_))
    i0 = idx[x_ == 1]
    i1 = idx[x_ == -1]
    idx = np.stack([i0,i1]).T
    if nmin is not None:
        duration = np.diff(idx, axis=-1).flatten()
        idx = idx[duration >= nmin]
    if sort:
        duration = np.diff(idx, axis=-1).flatten()
        idx = idx[np.argsort(duration)[::-1]]
    return idx


def fill_gaps(x, gap=None, fill=1):
    """
    Fills zero gaps in 1d-array.
    
    Parameters
    ----------
    x : ndarray
        1D array of non-negative numeric values
    gap : int or None, default None
        Max gap duration (if None - fill all gaps)
    fill : float, default 1
        Value to fill zeros

    Returns
    -------
    ndarray
        1D array - arrray with all gaps <= gap filled

    """
    assert x.ndim == 1 and not any(x < 0)
    x_ = np.copy(x)
    if gap is None:
        x_[~(x>0)] = fill
    elif gap > 0:
        x__ = (np.nan_to_num(x) == 0).astype(int)
        idx = find_intervals(x__)
        duration = np.diff(idx, axis=-1).flatten()
        idx = idx[duration <= gap]
        for i0, i1 in idx:
            x_[i0:i1] = fill
    return x_


def unique_sorted(x, return_dict=False):
    """
    Sorts unique values in descending order.
    
    Parameters
    ----------
    x : ndarray
        1D array of numeric values
    return_dict : bool, default False
        If False - return value, count arrays
        If True - return dict value -> count

    Returns
    -------
    value : ndarray, optional
        Unique values array (if return_dict is False)
    count : ndarray, optional
        Unique values counts (if return_dict is False)
    dict : dict, optional
        Dictionary unique value -> count (if return_dict is True)

    """
    x_ = x
    if isinstance(x, list):
        if isinstance(x[0], list):
            x_ = list(itertools.chain(*x))
            x_ = np.array(x_)
        else:
            x_ = np.array(x)
    x_ = x_[np.isfinite(x_)] if np.issubdtype(x_.dtype, float) else x_
    value, count = np.unique(x_, return_counts=True)
    idx = np.argsort(count)[::-1]
    value = value[idx]
    count = count[idx]
    if return_dict:
        return dict(zip(value, count))
    return value, count


def columns_to_datetime(df, tstart, tend=None, tz=None):
    """
    Converts DataFrame date/time columns to pandas.Timestamp.

    Parameters
    ----------
    df : DataFrame
        DataFrame
    columns : array_like
        List of columns to convert into datetime
    tz_col : str, optional
        Time zone column

    Returns
    -------
    DataFrame
        DataFrame with selected columns converted to pandas.Timestamp

    """
    def localize(row):
        fmt = "%Y-%m-%d %H:%M:%S"
        t = row[col].tz_localize("UTC").tz_convert(row[tz_col])
        t = pd.to_datetime(t.strftime(fmt), format=fmt)
        return t

    columns = find_columns_by_key(df, tstart) + find_columns_by_key(df, tend)
    tz_col = find_columns_by_key(df, tz)
    tz_col = tz_col[0] if len(tz_col) > 0 else None
    # if tz_col is not None:
    #     print("TZ", df[tz_col])
    for col in columns:
        if np.issubdtype(df[col].dtype, np.datetime64):
            continue
        if np.issubdtype(df[col].dtype, np.integer):
            df[col] = pd.to_datetime(df[col], unit="ms").dt.tz_localize(None)
        else:
            sample = df[col].values[0]
            fmt = "%Y-%m-%d %H:%M:%S.%f" if "-" in sample else "%m/%d/%y %H:%M:%S"
            if len(sample) > 5 and sample[-2:] in ["AM", "PM"]:
                 fmt = f"{fmt[:8]} %I:%M:%S %p"
            try:
                df[col] = pd.to_datetime(df[col], format=fmt).dt.tz_localize(None)
            except:
                df[col] = pd.to_datetime(df[col]).dt.tz_localize(None)
        if tz_col is not None:
            df[col] = df.apply(localize, axis=1)            
    return df


def from_ordinal(date, fmt="%Y-%m-%d"):
    """
    Converts ordinal day(s) to date(s), where day 1 = Jan 1st, 1 AD.

    Parameters
    ----------
    date : array_like or int
        Day(s) of type `int` (ordinal)
    fmt : str, default "%Y-%m-%d"
        Output date format

    Returns
    -------
    ndarray or str
        Date(s)

    """

    if isinstance(date, (list, np.ndarray)):
        date = np.array([from_ordinal(d) for d in date])
    elif not isinstance(date, str):
        date = datetime.datetime.fromordinal(date).strftime(fmt)
    return date


def to_ordinal(date):
    """
    Converts date(s) to ordinal day(s), where day 1 = Jan 1st, 1 AD.
    
    Parameters
    ----------
    date : array_like or str
        Date(s) of type `str`

    Returns
    -------
    ndarray or int
        Ordinal day(s)

    """
    if isinstance(date, (list, np.ndarray)):
        date = np.array(date)
        if np.issubdtype(date.dtype, str):
            try:
                date = pd.to_datetime(pd.Series(date))
                date = date.apply(pd.Timestamp.toordinal).values
            except:
                date = np.array([to_ordinal(d) for d in date])
    elif isinstance(date, str):
        date = datetime.datetime.strptime(date[:10], "%Y-%m-%d").toordinal()
    return date


def to_ordinal_day(date):
    """
    Converts date(s) to ordinal day(s), where day 1 = Jan 1st, 1 AD.
    
    Parameters
    ----------
    date : array_like or str
        Date(s) of type `str`

    Returns
    -------
    ndarray or int
        Ordinal day(s)

    """
    return to_ordinal(date)


def to_ordinal_week(date):
    """
    Converts date(s) to ordinal week(s), where week 1 = Jan 1st-7th, 1 AD.
    
    Parameters
    ----------
    date : array_like or str
        Date(s) of type `str`

    Returns
    -------
    ndarray or int
        Ordinal week(s)

    """
    return (to_ordinal(date) + 6) // 7


def to_ordinal_month(date):
    """
    Converts date(s) to ordinal month(s), where month 1 = Jan 1st-31st, 1 AD.
    
    Parameters
    ----------
    date : array_like or str
        Date(s) of type `str`

    Returns
    -------
    ndarray or int
        Ordinal month(s)

    """
    if isinstance(date, (list, np.ndarray)):
        month = np.array([ordinal_month(d) for d in date])
    else:
        if isinstance(date, str):
            date = datetime.datetime.strptime(date[:10], "%Y-%m-%d")
        else:
            date = datetime.datetime.fromordinal(date)
        month = date.month + 12 * (date.year - 1)
    return month


def to_ordinal_year(date):
    """
    Converts date(s) to ordinal year(s).
    
    Parameters
    ----------
    date : array_like or str
        Date(s) of type `str`

    Returns
    -------
    ndarray or int
        Ordinal year(s)

    """
    if isinstance(date, (list, np.ndarray)):
        year = np.array([ordinal_year(d) for d in date])
    elif isinstance(date, str):
        year = datetime.datetime.strptime(date[:10], "%Y-%m-%d").year
    else:
        year = datetime.datetime.fromordinal(date).year
    return year


def to_year_month_day(date):
    """
    Converts date(s) to (arrays of) year, month, day.
    
    Parameters
    ----------
    date : array_like or str
        Date(s) of type `str`

    Returns
    -------
    year : ndarray or int
        Year(s)
    month : ndarray or int
        Month(s) of the year
    day : ndarray or int
        Day(s) of the month

    """
    if isinstance(date, (list, np.ndarray)):
        date = np.array([to_year_month_day(d) for d in date]).T
    else:
        if not isinstance(date, str):
            date = datetime.datetime.fromordinal(date)
            date = date.strftime("%Y-%m-%d")
        date = np.array(date.split("-")).astype(int)
    year, month, day = list(date)
    return year, month, day


def to_range(date, pad_to_full_week=True):
    """
    Encloses date(s) into a continuos range of ordinal dates.
    
    Parameters
    ----------
    date : array_like
        List of dates of type `int` (ordinal) or `str`
    pad_to_full_week : bool, default True
        If True - pad range to full weeks so that its length % 7 == 0

    Returns
    -------
    ndarray
        Continuous range of ordinal days

    """
    try:
        idate = to_ordinal(date)
        i0 = idate.min()
        i1 = idate.max()
        if pad_to_full_week:
            i0 = i0 - to_weekdayiso(i0) + 1
            i1 = i1 - to_weekdayiso(i1) + 7
        idate = np.arange(i0, i1 + 1)
    except ValueError:
        idate = np.array([])
    return idate


def to_weekdayiso(date):
    """
    Converts date(s) to int day of the week (1 - Monday, 7 - Sunday).
    
    Parameters
    ----------
    date : array_like or str or int
        Date(s)

    Returns
    -------
    ndarray or int
        Day(s) of the week

    """
    idate = to_ordinal(date)
    weekday = (idate + 6) % 7 + 1
    return weekday


def to_weekdayiso_name(date):
    """
    Converts date(s) to name(s) of day of the week (Monday - Sunday).
    
    Parameters
    ----------
    date : array_like or str or int
        Date(s)

    Returns
    -------
    ndarray or str
        Day(s) of the week

    """
    d = list(calendar.day_name)
    d = dict(zip(np.arange(len(d)), np.array(d)))
    i = (to_ordinal(date) - 1) % 7
    w = np.vectorize(d.get)(i) if isinstance(i, np.ndarray) else d[i]
    return w


def to_weekdayiso_abbr(date):
    """
    Converts date(s) to name(s) of day of the week (Mon - Sun).
    
    Parameters
    ----------
    date : array_like or str or int
        Date(s)

    Returns
    -------
    ndarray or str
        Day(s) of the week

    """
    d = list(calendar.day_abbr)
    d = dict(zip(np.arange(len(d)), np.array(d)))
    i = (to_ordinal(date) - 1) % 7
    w = np.vectorize(d.get)(i) if isinstance(i, np.ndarray) else d[i]
    return w


def to_month_name(date):
    """
    Converts date(s) to name(s) of month (January - December).
    
    Parameters
    ----------
    date : array_like or str or int
        Date(s)

    Returns
    -------
    ndarray or str
        Month(s)

    """
    d = list(calendar.month_name)
    d = dict(zip(np.arange(len(d)), np.array(d)))
    i = (to_ordinal_month(date) - 1) % 12 + 1
    m = np.vectorize(d.get)(i) if isinstance(i, np.ndarray) else d[i]
    return m


def to_month_abbr(date):
    """
    Converts date(s) to name(s) of month (Jan - Dec).
    
    Parameters
    ----------
    date : array_like or str or int
        Date(s)

    Returns
    -------
    ndarray or str
        Month(s)

    """
    d = list(calendar.month_abbr)
    d = dict(zip(np.arange(len(d)), np.array(d)))
    i = (to_ordinal_month(date) - 1) % 12 + 1
    m = np.vectorize(d.get)(i) if isinstance(i, np.ndarray) else d[i]
    return m


def timezone_txt_to_minutes(tz):
    """
    Converts timezone name to minutes relative to UTC


    Parameters
    ----------
    tz : str
        Timezone, e.g. "Europe/Madrid" or "UTC+0100"

    Returns
    -------
    int
        Minutes

    """
    if isinstance(tz, (list, np.ndarray)):
        dt = np.array([timezone_txt_to_minutes(t) for t in tz]).T
    else:
        fmt = "%Y-%m-%d %H:%M:%S"
        t0 = datetime.datetime.strptime("2000-01-01 00:00:00+0000", fmt + "%z")
        try:
            try:
                t1 = t0.astimezone(pytz.timezone(tz)).strftime(fmt)
            except:
                t1 = t0.astimezone(pytz.timezone("UTC")).strftime(fmt)
                t0 = "2000-01-01 00:00:00" + re.sub("[a-zA-ZÀ-ž]", "", tz)
                t0 = datetime.datetime.strptime(t0, fmt + "%z")
            t0 = t0.astimezone(pytz.timezone("UTC")).strftime(fmt)
            t1 = datetime.datetime.strptime(t1, fmt)
            t0 = datetime.datetime.strptime(t0, fmt)
            dt = (t1 - t0).total_seconds() // 60
        except:
            dt = np.nan
    return np.round(dt)


def xticks_hours(dt=1, mode="24H", ax=None, **kwargs):
    """
    Matplotlib xticks as hours, assuming xlim is (0,1440) [min/day].
    
    Parameters
    ----------
    dt : int, default 1
        stride, hours
    mode : {"24H", "12H"}, default "24H"
        Time format
    ax : matplotlib.pyplot.Axes object, default None
        Axes for plotting
    **kwargs
        Keyword arguments

    Returns
    -------
    ax : matplotlib.pyplot.Axes object
        Axes for plotting

    """
    if ax is None:
        ax = plt.gca()
    t = np.arange(25)[::dt]
    h = t.astype(str)
    if mode.upper() == "12H":
        h = [datetime.datetime.strptime(f"{t_}", "%H").strftime('%I %p').upper() for t_ in t % 24]
    ax.set_xticks(t * 60, h, **kwargs)
    return ax


def xticks_days(x, ax=None, **kwargs):
    """
    Matplotlib xticks as days, assuming xlim is (0, 1440 x N days) [min].
    
    Parameters
    ----------
    x : array_like
        Data, to infer number of days
    ax : matplotlib.pyplot.Axes object, default None
        Axes for plotting
    **kwargs
        Keyword arguments
        
    Returns
    -------
    ax : matplotlib.pyplot.Axes object
        Axes for plotting

    """
    if ax is None:
        ax = plt.gca()
    n = len(x) // 1440
    t = np.arange(n + 1)
    ax.set_xticks(t * 1440, t, **kwargs)
    return ax


def xticks_dates(idate, mode="day", ax=None, **kwargs):
    """
    Matplotlib xticks as date(s).
    
    Parameters
    ----------
    idate : array_like
        Array or list of dates of type `int` (day 1 = Jan 1st, 1 AD)
    mode : {"day", "week", "fortnight", "month"}, default "day"
        Date spacing
    ax : matplotlib.pyplot.Axes object, default None
        Axes for plotting
    **kwargs
        Keyword arguments
        
    Returns
    -------
    ax : matplotlib.pyplot.Axes object
        Axes for plotting

    """
    assert mode in ["day", "week", "fortnight", "month", "year", 
                    "día", "semana", "quincena", "mes", "año"]
    if ax is None:
        ax = plt.gca()
    t = np.copy(idate)
    if mode in ["año", "year"]:
        _, month, day = to_year_month_day(t)
        t = t[(day == 1) & (month == 1)]
    if mode in ["mes", "month"]:
        day = to_year_month_day(t)[-1]
        t = t[day == 1]
    if mode in ["quincena", "fortnight"]:
        day = to_year_month_day(t)[-1]
        t = t[(day == 1) | (day == 15)]
    if mode in ["semana", "week"]:
        day = to_weekdayiso(t)
        t = t[day == 1]
    date = from_ordinal(t)
    if "rotation" not in kwargs:
        kwargs["rotation"] = 45
    if "ha" not in kwargs:
        kwargs["ha"] = "right"
    ax.set_xticks(t, date, **kwargs)
    return ax


def smoother(x, window, pad=np.nan, epsilon=1e-10, roll=False):
    """
    Smooth data using running Hann window.

    Parameters
    ----------
    x : ndarray
        Array of equispaced time series data
    window : int,
        Window size
    pad : float, default np.nan
        Value to pad x if roll is False
    epsilon : float, default 1e-10
        Cutoff to account values as non-zeros
    roll : bool, default False
        If True, pad with rolled x, else pad with zeros

    Returns
    -------
    ndarray
        Smoothed time series data

    """
    assert window < x.size
    hann = np.hanning(window)
    x_ = x.flatten()
    n = int(window) // 2
    pad0 = x_[-n:] if roll else np.ones((n)) * pad
    pad1 = x_[:n] if roll else np.ones((n)) * pad
    x_ = np.concatenate([pad0, x_, pad1])
    if np.all(np.isfinite(x)):
        w = np.correlate(x_, hann, mode='same') / np.sum(hann)
        w = w[n:-n].reshape(x.shape)
    else:
        w = np.lib.stride_tricks.sliding_window_view(x_, window)
        s = np.nansum((np.isfinite(w) * hann), 1)
        w = np.nansum(w * hann, 1) / s
        w = w[:x.size].reshape(x.shape)
    w[np.abs(w) < epsilon] = 0
    return w


def window_avg_std(t, x, window=14, smooth=0):
    """
    Calculates running window average and std.

    Parameters
    ----------
    t : ndarray
        1D array of time indices of values
    x : ndarray
        1D array of values
    window : int, default 14
        Window size
    smooth : int, default 0
        Window size to smooth average and std


    Returns
    -------
    t_avg : ndarray
        1D array of time indices of avg / std values
    x_avg : ndarray
        1D array of running window average values
    x_std : ndarray
        1D array of running window std values

    """
    n = t[-1] - t[0] + 1 + 2 * window
    pad = np.zeros((window)) * np.nan
    x_ = np.concatenate([pad, x, pad])
    t_avg = np.arange(n) + t[0] - window
    x_avg = np.zeros((n)) * np.nan
    x_std = np.zeros((n)) * np.nan
    for i in range(n):
        x_avg[i] = np.nanmean(x_[i-window//2:i+window//2+1])
        x_std[i] = np.nanstd(x_[i-window//2:i+window//2+1])
    if smooth:
        x_avg = smoother(x_avg, window=smooth)[window//2:][:n-2*window]
        x_std = smoother(x_std, window=smooth)[window//2:][:n-2*window]
    t_avg = t_avg[window:][:n-2*window]
    x_avg = x_avg[window:][:n-2*window]
    x_std = x_std[window:][:n-2*window]
    return t_avg, x_avg, x_std


def _remove_nonlocal_peaks(x, idx, window):
    """
    Removes redundant and non-local (edge) peaks.

    Parameters
    ----------
    x : ndarray
        Array of equispaced time series data
    idx : ndarray
        Indices of peaks
    window : int,
        Window size

    Returns
    -------
    ndarray
        Indices of true local peaks

    """
    def is_max(val, arr):
        return val == np.nanmax(arr) and val > np.nanmin(arr)
    idxs = []
    x_ = x.flatten()
    n = len(x_)
    for i in np.unique(idx):
        i0 = max(0, i - window // 2)
        i1 = min(n, i + window // 2)
        if len(idxs) > 0 and idxs[-1] > i - window // 2:
            continue
        if (i > 0) & (i < n) & is_max(x_[i], x_[i0:i+1]) & is_max(x_[i], x_[i:i1]):
            idxs.append(i)
    idxs = np.array(idxs).astype(int)
    return idxs


def series_peaks(x, window, smooth=False):
    """
    Finds local peaks in array of time series.

    Notes
    -----
    - NaN, Inf values NOT allowed

    Parameters
    ----------
    x : ndarray
        Array of equispaced time series data
    window : int
        Window size
    smooth : bool, default False
        If True - apply Hann window averaging smooth

    Returns
    -------
    idx : ndarray
        1D array of peak coordinate(s)
    score : ndarray
        1D array of peak height(s)

    """
    s = smoother(x, window) if smooth else x
    w = np.lib.stride_tricks.sliding_window_view(s, window)[::window//2]
    m, n = w.shape
    idx = np.argmax(w, axis=1)
    idx = idx + np.arange(len(w)) * (window // 2)
    idx = idx[(idx >= 0) & (idx < len(x))]
    idx = _remove_nonlocal_peaks(s, idx, window)
    score = s[idx]
    return idx, score


def histogram_peaks(x, bins=100, smooth=False):
    """
    Finds local peaks in histogram of values.

    Parameters
    ----------
    x : ndarray
        Array of data values
    bins : int or ndarray, default 100
        Histogram bins or number of bins
    smooth : bool, default False
        If True - apply Hann window averaging smooth

    Returns
    -------
    idx : ndarray
        1D array of peak coordinate(s)
    score : ndarray
        1D array of peak height(s)

    """
    if isinstance(bins, int):
        x1 = np.nanmin(x)
        x2 = np.nanmax(x)
        bins = np.linspace(x1, x2, bins)
    dx = np.diff(bins)[0]
    x_std = np.nanstd(x)
    window = int(x_std / dx)
    s = np.histogram(x, bins)[0]
    idx, score = series_peaks(s, window, smooth)
    idx = bins[idx]
    return idx, score


def window_sigmoid(n, m=None):
    """
    Generates sigmoid window (-1 to 1) of length n and width m.

    Parameters
    ----------
    n : int
        Window length
    m : int or None, default None
        Window width, if None, m = int(max(1, 0.2 * n))

    Returns
    -------
    ndarray
        1D array of sigmoid window

    """
    n = int(n)
    m = int(max(1, 0.2*n)) if m is None else int(m)
    t = np.arange(n) - (n-1) / 2
    w = 1 - 2 / (1 + np.exp(t / m))
    return np.round(w,2)


def window_boxcar(n, m=None):
    """
    Generates boxcar window (-1 to 1) of length n and width m.

    Parameters
    ----------
    n : int
        Window length
    m : int or None, default None
        Window width, if None, m = int(max(1, 0.2 * n))

    Returns
    -------
    ndarray
        1D array of boxcar window

    """
    n = int(n)
    m = int(max(1, 0.2*n)) if m is None else int(m)
    l = (n - m + (n + m) % 2) // 2
    w = -1 * np.ones((n))
    w[l:-l] = 1
    return w


def calc_covariance(x, window):
    """
    Calculates covariance with running window.

    Parameters
    ----------
    x : ndarray
        1D array of data points
    window : ndarray
        1D window array

    Returns
    -------
    ndarray
        1D array of covariance

    """
    n = len(window)
    pad = np.zeros((n//2)) * np.nan
    x_ = np.concatenate([pad, x, pad])
    x_ = np.lib.stride_tricks.sliding_window_view(x_, n).T
    x_ = (x_ - np.nanmean(x_, 0))
    cov = np.nanmean(x_[window > 0], 0) 
    cov = cov - np.nanmean(x_[window < 0], 0)
    n0 = np.sum(np.isfinite(x_[window < 0]),0)
    n1 = np.sum(np.isfinite(x_[window > 0]),0)
    mask = (n0 >= np.sum(window < 0) * 2 / 3) & (n1 >= np.sum(window > 0) * 2 / 3)
    cov[~mask] = np.nan
    cov = cov[:x.size]
    return cov


def calc_interpolation(x):
    """
    Linearly interpolates data array.

    Parameters
    ----------
    x : ndarray
        1D array of data points

    Returns
    -------
    ndarray
        1D array of interpolated data

    """
    x_ = np.zeros_like(x)
    mask = np.isfinite(x)
    if np.any(mask):
        x0 = x[mask][0]
        x1 = x[mask][-1]
        t = np.arange(len(x))
        x_ = np.interp(t, t[mask], x[mask], x0, x1)
    return x_


def anomaly_detection(x, wlen=10, wtype="step", cutoff=None):
    """
    Detects anomalies using running step or boxcar window.

    Parameters
    ----------
    x : ndarray
        1D array of data points
    wlen : int, default 10
        Window length
    wtype : {"step", "box"}, default "box"
        Window type
    cutoff : float or None, default None
        Cuoff scale for std

    Returns
    -------
    ndarray
        1D array of anomaly indices

    """
    if x.size < wlen:
        idx = np.array([])
    else:
        wlen = int(wlen)
        wfunc = {"step": window_sigmoid, "box": window_boxcar}
        window = wfunc[wtype](wlen)
        cov = calc_covariance(x, window)
        cov = calc_interpolation(cov)
        idx = series_peaks(cov, int(wlen * 2 / 3))
        std = np.nanstd(cov)
        if cutoff is not None:
            idx = idx[cov[idx] >= std * cutoff]
    return idx


def calc_cadence(steps):
    """
    Calculates walking and running cadence (steps/min).

    Parameters
    ----------
    steps : ndarray
        Array of equispaced time series data

    Returns
    -------
    walk : float
        Walking cadence
    run : float
        Running cadence

    """
    # histogram of steps
    bins = np.linspace(0,255,256)
    hist = np.histogram(steps, bins)[0]
    hist[0] = 0
    hist = np.log(hist+1)
    # log-linear fit in the scale-invariant range 20-60 steps/min
    x, y = bins[:-1], hist
    mask = (x >= 20) & (x <= 60)
    p = np.polyfit(x[mask], y[mask], 1)
    pred = np.polyval(p, x)
    pred[pred < 0] = 0
    # subtract scale-invariant trend to contrast peaks
    diff = hist - pred
    diff[diff < 0] = 0
    # find local maxima
    idx = series_peaks(diff, window=40)
    walk, run = np.nan, np.nan
    if len(idx):
        walk = idx[np.argmin(np.abs(idx - 110))]
        run = idx[np.argmin(np.abs(idx - 160))]
        run = run if run > walk else np.nan
        walk = walk if walk > 60 else np.nan
    return walk, run


def impute_bpm(bpm, tol=15):
    """
    Impputes short bpm gaps (defaukt 15 min) by linear interpolation.
    Some devices output bpm every 5 or 10 min. This function imputes 
    short gaps to make them compatible with bpm output every 1 min.

    Parameters
    ----------
    bpm : ndarray
        Array of equispaced time series data (N days x 1440 min)
    tol : int, default 15
        Max length of imputted intervals [minutes]

    Returns
    -------
    ndarray
        Imputed bpm of the same shape as the input bpm

    """

    # flatten to 1D and make placeholder for imputted bpm
    x = bpm.flatten()
    ibpm = np.copy(x)
    # Collect intervals id that should not be interpolated (longer then tol)
    idx = find_intervals(x <= 0, tol=0, nmin=tol)
    # Mask non-zero bpm as reference points for interpolation
    mask = x > 0
    # make interpolation, then turn too long gaps back to zero (using idx)
    if np.any(mask):
        t = np.arange(len(x))
        ibpm = np.interp(t, t[mask], x[mask])
        for i0, i1, in idx:
            ibpm[i0:i1] = 0
    # reshape flattened 1D array back to the shape of the input array
    ibpm = ibpm.reshape(bpm.shape)
    return ibpm


def defragment_sleep(sleep, tol=60):
    """
    Defragments sleep intervals. Some samples have fragmented sleep records.
    This function imputes short gaps (defalut 60 min) to merge a series of 
    fragmented sleep bouts into a countinous sleep interval.

    Parameters
    ----------
    sleep : ndarray
        Array of equispaced time series data (N days x 1440 min)
    tol : int, default 60
        Max length of imputted intervals [minutes]

    Returns
    -------
    ndarray
        Defragmented sleep of the same shape as the input sleep

    """

    # flatten to 1D and make placeholder for imputted sleep
    x = sleep.flatten()
    dsleep = np.zeros_like(x)
    # Collect continuous sleep intervals indices and impute with ones
    idxs = find_intervals(x > 0, tol=tol)
    for i0, i1 in idxs:
        dsleep[i0:i1] = 1
    # reshape flattened 1D array back to the shape of the input array
    dsleep = dsleep.reshape(sleep.shape)
    return dsleep


import types
__all__ = [name for name, thing in globals().items()
          if not (name.startswith('_') or isinstance(thing, types.ModuleType))]
del types

